class I2C:

    def __init__(self, *, frequency=400000):
        from adafruit_blinka.microcontroller.nova import Connection
        self._nova = Connection.getInstance()
        self._nova.setNumericalBase(10)
        self._nova.setOperationMode(0, "I2C")
        self._nova.setPullUpStateI2C(0, "EN")
        self._nova.setClockI2C(0, frequency)

    def scan(self):

        scanResults = []

        for i in range(8, 121):
            result = self._nova.scanAddrI2C(0, i<<1)

            resp = result.split(" ")

            if resp[3] == 'OK':
                scanResults.append(i)

        return scanResults

    def writeto(self, address, buffer, *, start=0, end=None, stop=True):

        end = end if end else len(buffer)

        self._nova.startI2C(0, address<<1)

        for i in range(start, end): 
            self._nova.writeByteI2C(0, buffer[i])

        if stop:
            self._nova.endI2C(0)
        else:
            self._nova.endI2C(0, True)

    def readfrom_into(self, address, buffer, *, start=0, end=None, stop=True):

        end = end if end else len(buffer)

        result = self._nova.readBytesI2C(0, address<<1, len(buffer[start:end]))

        if result != "-NG":
            resp = result.split(" ")

            for i in range(len(buffer[start:end])):
                buffer[start+i] = int(resp[2+i])
        else:
            raise RuntimeError("Received error response from Binho Nova, result = " + result)

    def writeto_then_readfrom(self, address, buffer_out, buffer_in, *,
                              out_start=0, out_end=None,
                              in_start=0, in_end=None, stop=False):
    
        out_end = out_end if out_end else len(buffer_out)
        in_end = in_end if in_end else len(buffer_in)

        self._nova.startI2C(0, address<<1)

        for i in range(out_start, out_end): 
            self._nova.writeByteI2C(0, buffer_out[i])

        self._nova.endI2C(0, True)

        result = self._nova.readBytesI2C(0, address<<1, len(buffer_in[in_start:in_end]))

        if result != "-NG":
            resp = result.split(" ")

            for i in range(len(buffer_in[in_start:in_end])):
                buffer_in[in_start+i] = int(resp[2+i])
        else:
            raise RuntimeError("Received error response from Binho Nova, result = " + result)



from adafruit_blinka.microcontroller.ft232h.pin import Pin

class I2C:

    def __init__(self, *, frequency=400000):
        # change GPIO controller to I2C
        from pyftdi.i2c import I2cController
        self._i2c = I2cController()
        self._i2c.configure('ftdi:///1', frequency=frequency)
        Pin.ft232h_gpio = self._i2c.get_gpio()

    def scan(self):
        return [addr for addr in range(0x79) if self._i2c.poll(addr)]

    def writeto(self, address, buffer, *, start=0, end=None, stop=True):
        end = end if end else len(buffer)
        port = self._i2c.get_port(address)
        port.write(buffer[start:end], relax=stop)

    def readfrom_into(self, address, buffer, *, start=0, end=None, stop=True):
        end = end if end else len(buffer)
        port = self._i2c.get_port(address)
        result = port.read(len(buffer[start:end]), relax=stop)
        for i, b in enumerate(result):
            buffer[start+i] = b

    def writeto_then_readfrom(self, address, buffer_out, buffer_in, *,
                              out_start=0, out_end=None,
                              in_start=0, in_end=None, stop=False):
        out_end = out_end if out_end else len(buffer_out)
        in_end = in_end if in_end else len(buffer_in)
        port = self._i2c.get_port(address)
        result = port.exchange(buffer_out[out_start:out_end],
                               in_end-in_start,
                               relax=True)
        for i, b in enumerate(result):
            buffer_in[in_start+i] = b



class SPI:
    MSB = 0
    PAYLOAD_MAX_LENGTH = 64

    def __init__(self, clock):
        from adafruit_blinka.microcontroller.nova import Connection
        self._nova = Connection.getInstance()
        self._nova.setNumericalBase(10)
        self._nova.setOperationMode(0, 'SPI')
        self._nova.setClockSPI(0, clock)
        self._nova.setModeSPI(0, 0)
        self._nova.setIOpinMode(0, 'DOUT')
        self._nova.setIOpinMode(1, 'DOUT')
        self._nova.beginSPI(0)

        # Cpol and Cpha set by mode
        # Mode  Cpol Cpha
        #  0     0    0
        #  1     0    1
        #  2     1    0
        #  3     1    1

    def init(self, baudrate=100000, polarity=0, phase=0, bits=8,
             firstbit=MSB, sck=None, mosi=None, miso=None):
        #print("baudrate: " + str(baudrate))
        #print("mode: " + str((polarity<<1) | (phase)))
        self._nova.setClockSPI(0, baudrate)
        self._nova.setModeSPI(0, (polarity<<1) | (phase))

    @staticmethod
    def get_received_data(lineOutput):
        return (lineOutput.split('RXD ')[1])

    @property
    def frequency(self):
        return self._nova.getClockSPI(0).split('CLK ')[1]

    def write(self, buf, start=0, end=None):
        end = end if end else len(buf)
        chunks, rest = divmod(end - start, self.PAYLOAD_MAX_LENGTH)
        for i in range(chunks):
            chunk_start = start + i * self.PAYLOAD_MAX_LENGTH
            chunk_end = chunk_start + self.PAYLOAD_MAX_LENGTH
            buffer_data = buf[chunk_start:chunk_end]
            self._nova.clearBuffer(0)
            self._nova.writeToBuffer(0, 0, buffer_data)
            self._nova.transferBufferSPI(0, chunk_end - chunk_start + 1)
        if rest:
            buffer_data = buf[-1*rest:]
            self._nova.clearBuffer(0)
            self._nova.writeToBuffer(0, 0, buffer_data)
            self._nova.transferBufferSPI(0, rest)

    def readinto(self, buf, start=0, end=None, write_value=0):
        end = end if end else len(buf)
        for i in range(start, end):
            buf[start+i] = int(self.get_received_data(self._nova.transferSPI(0, write_value)))

    def write_readinto(self, buffer_out, buffer_in, out_start=0, out_end=None, in_start=0, in_end=None):
        out_end = out_end if out_end else len(buffer_out)
        in_end = in_end if in_end else len(buffer_in)
        readlen = in_end-in_start
        writelen = out_end-out_start
        if readlen > writelen:
            # resize out and pad with 0's
            tmp = bytearray(buffer_out)
            tmp.extend([0] * (readlen - len(buffer_out)))
            buffer_out = tmp
        i = 0
        for data_out in buffer_out:
            data_in = int(self.get_received_data(self._nova.transferSPI(0, data_out)))
            if i < readlen:
                buffer_in[in_start+i] = data_in
            i += 1



"""Platform agnostic time implementation"""

from adafruit_blinka import agnostic

# We intentionally are patching into this namespace so skip the wildcard check.
# pylint: disable=unused-wildcard-import,wildcard-import

if agnostic.implementation == "circuitpython":
    from time import *
elif agnostic.implementation == "micropython":
    import utime
    from utime import sleep

    from ucollections import namedtuple
    _struct_time = namedtuple("struct_time", ("tm_year", "tm_mon", "tm_mday",
                                              "tm_hour", "tm_min", "tm_sec",
                                              "tm_wday", "tm_yday", "tm_isdst"))

    #pylint: disable=too-many-arguments
    def _marshal_time(tm_year, tm_mon, tm_mday, tm_hour=0, tm_min=0, tm_sec=0, tm_wday=-1,
                      tm_yday=-1, tm_isdst=-1):
        """Construct struct_time with default values."""
        _struct_time(tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_wday, tm_yday, tm_isdst)

    def struct_time(time_tuple):
        """Create a struct_time"""
        return _marshal_time(*time_tuple)

    #pylint: disable=invalid-name
    _total_ms = 0
    _prev_ticks_ms = utime.ticks_ms()
    def monotonic():
        """A monotonically increasing time in seconds. No defined start time."""
        # Assumes that monotonic is called more frequently than the wraparound of micropython's
        # utime.ticks_ms()
        global _prev_ticks_ms, _total_ms #pylint: disable=global-statement
        ticks_ms = utime.ticks_ms()
        _total_ms += utime.ticks_diff(ticks_ms, _prev_ticks_ms)
        _prev_ticks_ms = ticks_ms
        return _total_ms * 0.001
